/*******************************************************************************
*
*	TITLE:		printf.c
*
*	VERSION:	1.2                          
*
*	DATE:		28-Dec-2004
*
*	AUTHORS:	Innovation First
*
*				R. Kevin Watson
*				kevinw@jpl.nasa.gov
*
*	COMMENTS:	This version of the printf() function can send output to either
*				of the serial ports by setting the value of the global variable
*				"stdout" before calling the printf() function. Setting the value 
*				to "SERIAL_PORT_ONE" will send the output to serial port one. 
*				Likewise, setting the value to "SERIAL_PORT_TWO" will send the 
*				output to serial port two. Setting the value to "NUL" will send
*				the output to the null device, meaning that the output is sent
*				nowhere. These values are #define'd in printf.h. As an example,
*
*				  stdout = SERIAL_PORT_ONE;
*				  printf("Kernighan ");
*				  stdout = NUL;
*				  printf("and ");
*				  stdout = SERIAL_PORT_TWO;
*				  printf("Ritchie");
*
*				will send the text "Kernighan" to the peripheral device attached
*				to serial port one, the text "Ritchie" to the device attached to 
*				serial port two and the text "and " won't be sent anywhere.
*				By default, output is sent to the null device, which is the only
*				output guaranteed to be present.
*
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	             1.0  Innovation First's original code
*	20-Dec-2004  1.1  RKW - Renamed and modified IFI's printf_lib.c/h files
*	                  to use the code in serial_ports.c
*	28-Dec-2004  1.2  RKW - Using preprocessor directives, added the ability 
*	                  to enable/disable individual receive and transmit code.
*	                  See comments within Write_Byte_To_Uart()'s header for
*	                  more information.
*	28-Dec-2004  1.2  RKW - Renamed "Current_Serial_Port" to a more unix-like 
*	                  "stdout". Updated documentation.
*	28-Dec-2004  1.2  RKW - Added the NUL device as an option for stdout.
*
*******************************************************************************/

#include "ifi_default.h"
#include "ifi_utilities.h"
#include "string.h"
#include "serial_ports.h"
#include "printf.h"

// the following should be enough for a 32 bit word
#define PRINT_BUF_LEN 20
#define PAD_RIGHT 1
#define PAD_ZERO  2

extern char ifi_printfBufr[]; // declared in ifi_utilities.c
rom char *nullStr = "(null)";

static char print_buf[PRINT_BUF_LEN];
static char scr[2];
static int  k;

// by default printf() output is sent to the null device,
// which the only device guaranteed to be present. 
unsigned char stdout = NUL;

/*******************************************************************************
*
*	FUNCTION:		Write_Byte_To_Uart()
*
*	PURPOSE:		Sends a byte of data to the serial port defined by stdout.
*
*	CALLED FROM:
*
*	PARAMETERS:		unsigned char
*
*	RETURNS:		Nothing
*
*	COMMENTS:		If ENABLE_SERIAL_PORT_ONE_TX and ENABLE_SERIAL_PORT_TWO_TX
*					have both been #define'd in serial_ports.h, the three
*					possible values for stdout are NUL, SERIAL_PORT_ONE_TX and
*					SERIAL_PORT_TWO_TX.
*
*					If only ENABLE_SERIAL_PORT_ONE_TX has been #define'd in
*					serial_ports.h, the two possible values for stdout are NUL,
*					and SERIAL_PORT_ONE.
*
*					If only ENABLE_SERIAL_PORT_TWO_TX has been #define'd in
*					serial_ports.h, the two possible values for stdout are NUL,
*					and SERIAL_PORT_TWO.
*
*					If ENABLE_SERIAL_PORT_ONE_TX and ENABLE_SERIAL_PORT_TWO_TX
*					are not #define'd in serial_ports.h, this function just
*					becomes a stub and output from printf() will be sent
*					to the bit bucket.
*
*******************************************************************************/
static void Write_Byte_To_Uart(unsigned char byte)
{
	if(stdout == NUL)
	{
		// send the data to the bit bucket
	}
	#ifdef ENABLE_SERIAL_PORT_ONE_TX
	else if(stdout == SERIAL_PORT_ONE)
	{
		// send the data to serial port one
		Write_Serial_Port_One(byte);
	}
	#endif
	#ifdef ENABLE_SERIAL_PORT_TWO_TX
	else if(stdout == SERIAL_PORT_TWO)
	{
		// send the data to serial port two
		Write_Serial_Port_Two(byte);
	}
	#endif
}

/*******************************************************************************
*
*	FUNCTION:		prints()
*
*	PURPOSE:		Pads the output stream.
*
*	CALLED FROM:
*
*	PARAMETERS:
*
*	RETURNS:		Nothing
*
*	COMMENTS:

*******************************************************************************/
static int prints(char *string, int width, int pad)
{
	register int pc = 0;
	register int padchar = ' ';

	if(width > 0)
	{
		register int len = 0;
		register char *ptr;

		for(ptr = string; *ptr; ++ptr)
			++len;
		if (len >= width)
			width = 0;
		else
			width -= len;
		if(pad & PAD_ZERO)
			padchar = '0';
	}
	
	if(!(pad & PAD_RIGHT))
	{
		for( ; width > 0; --width)
		{
			Write_Byte_To_Uart(padchar);
			++pc;
		}
	}
	
	for( ; *string ; ++string)
	{
		Write_Byte_To_Uart(*string);
		++pc;
	}
	
	for( ; width > 0; --width)
	{
		Write_Byte_To_Uart(padchar);
		++pc;
	}
	
	return pc;
}

/*******************************************************************************
*
*	FUNCTION:		printi()
*
*	PURPOSE:		Converts the output stream to the proper width and base.
*
*	CALLED FROM:
*
*	PARAMETERS:
*
*	RETURNS:		Nothing
*
*	COMMENTS:

*******************************************************************************/
static int printi(int i, int b, int sg, int width, int pad, int letbase)
{
	register char *s;
	register int t;
	register int neg = 0;
	register int pc = 0;
	register unsigned int u = i;

	print_buf[0] = 0xBE;
	print_buf[PRINT_BUF_LEN] = 0xEF;

	if(i == 0)
	{
		print_buf[0] = '0';
		print_buf[1] = '\0';
		return prints(print_buf, width, pad);
	}

	if(sg && b == 10 && i < 0)
	{
		neg = 1;
		u = -i;
	}

	s = print_buf + PRINT_BUF_LEN-1;
	*s = '\0';

	while(u)
	{
		t = u % b;
		if(t >= 10)
		{
			t += letbase - '0' - 10;
		}
		*--s = t + '0';
		u /= b;
	}

	if(neg)
	{
		if(width && (pad & PAD_ZERO))
		{
			Write_Byte_To_Uart('-');
			++pc;
			--width;
		}
		else
		{
			*--s = '-';
		}
	}

	return pc + prints (s, width, pad);
}

/*******************************************************************************
*
*	FUNCTION:		print()
*
*	PURPOSE:		Parses the output stream.
*
*	CALLED FROM:
*
*	PARAMETERS:
*
*	RETURNS:
*
*	COMMENTS:

*******************************************************************************/
static int print(char *format, int *varg)
{
	char tmpBufr[40];
	register int width;
	register int pad;
	register int pc = 0;

	for(; *format != 0; ++format)
	{
		if(*format == '%')
		{
			++format;
			width = 0;
			pad = 0;

			if(*format == '\0')
			{
				break;
			}

			if(*format == '%')
			{
				goto out;
			}

			if(*format == '-')
			{
				++format;
				pad = PAD_RIGHT;
			}
			
			while(*format == '0')
			{
				++format;
				pad |= PAD_ZERO;
			}

			for( ; *format >= '0' && *format <= '9'; ++format)
			{
				width *= 10;
				width += *format - '0';
			}

			if(*format == 's') 
      		{
        		strcpypgm2ram(tmpBufr,(rom char *) *varg--);

        		if (tmpBufr[0] == 0)
				{
          			strcpypgm2ram(tmpBufr,nullStr);
				}
				
				pc += prints (tmpBufr, width, pad);
				continue;
			}

			if(*format == 'd')
			{
				pc += printi(*varg--, 10, 1, width, pad, 'a');
				continue;
			}

			if(*format == 'x')
			{
				pc += printi(*varg--, 16, 0, width, pad, 'a');
				continue;
			}

			if(*format == 'X')
			{
				pc += printi(*varg--, 16, 0, width, pad, 'A');
				continue;
			}

			if(*format == 'u')
			{
				pc += printi(*varg--, 10, 0, width, pad, 'a');
				continue;
			}

			if(*format == 'l')  // assumes lx
			{
				pc += printi(*varg--, 16, 0, width, pad, 'a');
				pc += printi(*varg--, 16, 0, width, pad, 'a');
				format++; // skip over x
				continue;
			}

			if(*format == 'b')
			{
				pc += printi(*varg--, 2, 0, width, 2, 'a');
				continue;
			}
		}
		else
		{
			out:
//  Commented out by RKW on 20 Dec 2004 because line feeds are 
//	needed for terminal emulation
//			if (*format == '\n') 
//				*format = '\r';  // replace line feed with cr
			Write_Byte_To_Uart (*format);
			++pc;
		}
	}
	return pc;
}

/*******************************************************************************
*
*	FUNCTION:		printf()
*
*	PURPOSE:		Formats an output stream.
*
*	CALLED FROM:
*
*	PARAMETERS:
*
*	RETURNS:
*
*	COMMENTS:
*
*******************************************************************************/
int printf(rom const char *format, ...)
{
    register int *varg = (int *)(&format);

	// Since constant strings are kept in program (flash) memory, 
	// the strcpypgm2ram routine copies the string from flash to ram. 
    strcpypgm2ram(ifi_printfBufr,(rom char *) format);

    varg--; //adjust stack for Microchip C Compiler
    return print(ifi_printfBufr, varg);
}

